/* matrix.c */

/* VIS-5D version 3.0 */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



#ifdef stellar
#  include <fastmath.h>
#else
#  include <math.h>
#endif


#include "matrix.h"



/*** make_matrix ******************************************************
   Make a transformation matrix to perform the given rotation, scale
   and rotation.  This function uses the fast matrix post-concatenation
   techniques from Graphics Gems.
**********************************************************************/
void make_matrix( rotx, roty, rotz, scale, transx, transy, transz, mat )
float  rotx, roty, rotz, scale, transx, transy, transz;
Matrix  mat;
{
   register float  sx, sy, sz, cx, cy, cz, t;
   register int i;

   /* Get sin and cosine values */
   sx = sin(rotx / 57.2957);
   cx = cos(rotx / 57.2957);
   sy = sin(roty / 57.2957);
   cy = cos(roty / 57.2957);
   sz = sin(rotz / 57.2957);
   cz = cos(rotz / 57.2957);

   /* X Rotation */
   mat[0][0] = 1.0;  mat[0][1] = 0.0;  mat[0][2] = 0.0;  mat[0][3] = 0.0;
   mat[1][0] = 0.0;  mat[1][1] =  cx;  mat[1][2] =  sx;  mat[1][3] = 0.0;
   mat[2][0] = 0.0;  mat[2][1] = -sx;  mat[2][2] =  cx;  mat[2][3] = 0.0;
   mat[3][0] = 0.0;  mat[3][1] = 0.0;  mat[3][2] = 0.0;  mat[3][3] = 1.0;

   /* Y Rotation */
   for (i=0;i<4;i++) {
      t = mat[i][0];
      mat[i][0] = mat[i][2]*sy + t*cy;
      mat[i][2] = mat[i][2]*cy - t*sy;
   }

   /* Z Rotation */
   for (i=0;i<4;i++) {
      t = mat[i][0];
      mat[i][0] = t*cz - mat[i][1]*sz;
      mat[i][1] = t*sz + mat[i][1]*cz;
   }

   /* Scale */
   for (i=0;i<3;i++) {
      mat[i][0] *= scale;
      mat[i][1] *= scale;
      mat[i][2] *= scale;
   }

   /* Translation */
   mat[3][0] = transx;
   mat[3][1] = transy;
   mat[3][2] = transz;

}




/*** mat_mul *************************************************************
   Multiply two matrices.
   r = a * b
*************************************************************************/
void mat_mul(r, a, b)
Matrix  r, a, b;
{
   register int i, j, k;
   Matrix temp;

   for (i=0; i<4; i++) {
      for (j=0; j<4; j++) {
         temp[i][j] = 0.0;
         for (k=0; k<4; k++)
            temp[i][j] += a[i][k] * b[k][j];
      }
   }
   mat_copy(r, temp);
}



/*** mat_concat **********************************************************
   Concatenate two matrices.
   r = a * b
*************************************************************************/
void mat_concat(r, a, b)
Matrix  r, a, b;
{
   Matrix temp;
   register int i, j, k;
   float tx, ty, tz;

   tx = a[3][0];  a[3][0] = -tx;
   ty = a[3][1];  a[3][1] = -ty;
   tz = a[3][2];  a[3][2] = -tz;

   for (i=0; i<4; i++) {
      for (j=0; j<4; j++) {
         temp[i][j] = 0.0;
         for (k=0; k<4; k++) 
            temp[i][j] += a[i][k] * b[k][j];
      }
   }
   temp[3][0] = tx;
   temp[3][1] = ty;
   temp[3][2] = tz;

   mat_copy(r, temp);
}




/*** mat_copy ***********************************************************
   Copy a matrix from src to dest.
************************************************************************/
void mat_copy( dest, src )
Matrix dest, src;
{
/*   memcpy( dest, src, 16*sizeof(float) );*/
   int i,j;

   for (i=0;i<4;i++)
      for (j=0;j<4;j++)
         dest[i][j] = src[i][j];


}




/*** mat_inv ***************************************************************
   Find the inverse of matrix 'mat' and return it in 'inv'.
   inv = mat ^ -1
***************************************************************************/
void mat_inv( inv, mat)
Matrix inv, mat;
{
   int i, j;
   float det, sub();

   mat_copy( inv, mat );
   det = mat[0][0] * sub(mat, 0, 0)
        - mat[1][0] * sub(mat, 1, 0)
        + mat[2][0] * sub(mat, 2, 0);
   for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
         inv[j][i] = (((i+j)%2) ? -1 : 1) * sub(mat, i, j) / det;
      }
   }
}



float sub(m, i, j)
float m[4][4];
int i, j;
{
   int i1, i2, j1, j2;

   i1 = (i==0) ? 1 : 0;
   i2 = (i==2) ? 1 : 2;
   j1 = (j==0) ? 1 : 0;
   j2 = (j==2) ? 1 : 2;
   return(m[i1][j1] * m[i2][j2] - m[i1][j2] * m[i2][j1]);
}




/*** vec_norm *********************************************************
   Normalize (scale) a vector, vin, to normal length (1.0).
   Return result in vout.
**********************************************************************/
void vec_norm( vin, vout )
float vin[3],vout[3];
{
   float len;

   len = sqrt( vin[0]*vin[0] + vin[1]*vin[1] + vin[2]*vin[2] );

   if (len!=0.0) {
      vout[0] = vin[0] / len;
      vout[1] = vin[1] / len;
      vout[2] = vin[2] / len;
   }
   else {
      vout[0] = vout[1] = 0.0;
      vout[2] = 1.0;
   }
}



/*** mat_vecmul **********************************************************
   Multiply a vector, v,  by a matrix, m.
*************************************************************************/
mat_vecmul(v, m)
float v[3];
Matrix m;
{
   float xp, yp, zp, wp;
   float vec4[4];
   register float *mp, *vp;
   register int i;

   vec4[0] = v[0];
   vec4[1] = v[1];
   vec4[2] = v[2];
   vec4[3] = 1.0;

   mp = (float *)m;
   vp = (float *)vec4;
   xp = yp = zp = wp = 0.0;
   for (i=0; i<4; i++) {
      xp += *mp++ * *vp;
      yp += *mp++ * *vp;
      zp += *mp++ * *vp;
      wp += *mp++ * *vp;
      vp++;
   }
   v[0] = xp / wp;
   v[1] = yp / wp;
   v[2] = zp / wp;
}


print_matrix( mat )
Matrix mat;
{
   int i,j;

   for (i=0;i<4;i++) {
      for (j=0;j<4;j++)
         printf("%f ", mat[i][j]);
      printf("\n");
    }
 }
